@@ -32,7 +32,7 @@ import warnings
 import ctypes
 from ctypes import c_char_p, POINTER, c_size_t, c_bool, c_void_p, c_int
 
-from matlab_wrapper.typeconv import np_to_mat
+from matlab_wrapper.typeconv import dtype_to_mat
 
 class mxArray(ctypes.Structure):
     pass
@@ -101,7 +101,7 @@ class MatlabSession(object):
                     break
 
         if matlab_root is None:
-            raise RuntimeError("MATLAB location is unknown (set matlab_root)")
+            raise RuntimeError("MATLAB location is unknown (try to initialize MatlabSession with matlab_root properly set)")
 
         self._matlab_root = matlab_root
 
@@ -164,26 +164,36 @@ class MatlabSession(object):
         self._libmx.mxGetClassName.argtypes = (POINTER(mxArray),)
         self._libmx.mxGetClassName.restype = c_char_p
 
-        self._libmx.mxIsNumeric.restype = c_bool
         self._libmx.mxIsNumeric.argtypes = (POINTER(mxArray),)
+        self._libmx.mxIsNumeric.restype = c_bool
+
+        self._libmx.mxIsCell.argtypes = (POINTER(mxArray),)
+        self._libmx.mxIsCell.restype = c_bool
 
-        self._libmx.mxIsComplex.restype = c_bool
         self._libmx.mxIsComplex.argtypes = (POINTER(mxArray),)
+        self._libmx.mxIsComplex.restype = c_bool
 
-        self._libmx.mxGetData.restype = POINTER(c_void_p)
         self._libmx.mxGetData.argtypes = (POINTER(mxArray),)
+        self._libmx.mxGetData.restype = POINTER(c_void_p)
         self._libmx.mxGetData.errcheck = error_check
 
-        self._libmx.mxGetImagData.restype = POINTER(c_void_p)
         self._libmx.mxGetImagData.argtypes = (POINTER(mxArray),)
+        self._libmx.mxGetImagData.restype = POINTER(c_void_p)
         self._libmx.mxGetImagData.errcheck = error_check
 
+        self._libmx.mxGetCell.argtypes = (POINTER(mxArray), c_size_t)
+        self._libmx.mxGetCell.restype = POINTER(mxArray)
+        self._libmx.mxGetCell.errcheck = error_check
+
+        self._libmx.mxSetCell.argtypes = (POINTER(mxArray), c_size_t, POINTER(mxArray))
+        self._libmx.mxSetCell.restype = None
+
         self._libmx.mxArrayToString.argtypes = (POINTER(mxArray),)
         self._libmx.mxArrayToString.restype = c_char_p
         self._libmx.mxArrayToString.errcheck = error_check
 
-        self._libmx.mxCreateString.restype = POINTER(mxArray)
         self._libmx.mxCreateString.argtypes = (c_char_p,)
+        self._libmx.mxCreateString.restype = POINTER(mxArray)
         self._libmx.mxCreateString.errcheck = error_check
 
         self._libmx.mxCreateNumericArray.argtypes = (c_size_t, POINTER(c_size_t), c_int, c_int)
@@ -194,11 +204,9 @@ class MatlabSession(object):
         self._libmx.mxCreateLogicalArray.restype = POINTER(mxArray)
         self._libmx.mxCreateLogicalArray.errcheck = error_check
 
-        self._libmx.mxGetData.argtypes = (POINTER(mxArray),)
-        self._libmx.mxGetData.restype = POINTER(c_void_p)
-
-        self._libmx.mxGetImagData.argtypes = (POINTER(mxArray),)
-        self._libmx.mxGetImagData.restype = POINTER(c_void_p)
+        self._libmx.mxCreateCellArray.argtypes = (c_size_t, POINTER(c_size_t))
+        self._libmx.mxCreateCellArray.restype = POINTER(mxArray)
+        self._libmx.mxCreateCellArray.errcheck = error_check
 
         self._libmx.mxDestroyArray.argtypes = (POINTER(mxArray),)
         self._libmx.mxDestroyArray.restype = None
@@ -281,6 +289,7 @@ class MatlabSession(object):
 
         """
         pm = self._libeng.engGetVariable(self._ep, name)
+
         ndims = self._libmx.mxGetNumberOfDimensions(pm)
         dims = self._libmx.mxGetDimensions(pm)
         numelems = self._libmx.mxGetNumberOfElements(pm)
@@ -291,59 +300,8 @@ class MatlabSession(object):
         data = self._libmx.mxGetData(pm)
         imag_data = self._libmx.mxGetImagData(pm)
 
-        if is_numeric:
-            datasize = numelems*elem_size
 
-            real_buffer = ctypes.create_string_buffer(datasize)
-            ctypes.memmove(real_buffer, data, datasize)
-            pyarray = np.ndarray(
-                buffer=real_buffer,
-                shape=dims[:ndims],
-                dtype=class_name,
-                order='F'
-            )
-
-            if is_complex:
-                imag_buffer = ctypes.create_string_buffer(datasize)
-                ctypes.memmove(imag_buffer, imag_data, datasize)
-                pyarray_imag = np.ndarray(
-                    buffer=imag_buffer,
-                    shape=dims[:ndims],
-                    dtype=class_name,
-                    order='F'
-                )
-
-                pyarray = pyarray + pyarray_imag * 1j
-
-            out = pyarray.squeeze()
-
-
-        elif class_name == 'char':
-            datasize = numelems + 1
-
-            pystring = ctypes.create_string_buffer(datasize+1)
-            self._libmx.mxGetString(pm, pystring, datasize)
-
-            out = pystring.value
-
-
-        elif class_name == 'logical':
-            datasize = numelems*elem_size
-
-            buf = ctypes.create_string_buffer(datasize)
-            ctypes.memmove(buf, data, datasize)
-
-            pyarray = np.ndarray(
-                buffer=buf,
-                shape=dims[:ndims],
-                dtype='bool',
-                order='F'
-            )
-
-            out = pyarray.squeeze()
-
-        else:
-            raise NotImplementedError('{}-arrays are not implemented'.format(class_name))
+        out = convert_mat_to_ndarray(self._libmx, pm)
 
 
         self._libmx.mxDestroyArray(pm)
@@ -357,59 +315,8 @@ class MatlabSession(object):
         """Put a variable to MATLAB workspace.
 
         """
-        if isinstance(value, str):
-            pm = self._libmx.mxCreateString(value)
-
-        elif isinstance(value, dict):
-            raise NotImplementedError('dicts are not supported.')
-
-        else:
-            value = np.array(value, ndmin=2)
-
-
-        if isinstance(value, np.ndarray) and value.dtype.kind in ['i','u','f','c']:
-            dim = value.ctypes.shape_as(c_size_t)
-            complex_flag = (value.dtype.kind == 'c')
-
-            pm = self._libmx.mxCreateNumericArray(
-                value.ndim,
-                dim,
-                np_to_mat(value.dtype),
-                complex_flag
-            )
-
-            mat_data = self._libmx.mxGetData(pm)
-            np_data = value.real.tostring('F')
-            ctypes.memmove(mat_data, np_data, len(np_data))
-
-            if complex_flag:
-                mat_data = self._libmx.mxGetImagData(pm)
-                np_data = value.imag.tostring('F')
-                ctypes.memmove(mat_data, np_data, len(np_data))
-
-        elif isinstance(value, np.ndarray) and value.dtype.kind == 'b':
-            dim = value.ctypes.shape_as(c_size_t)
-
-            pm = self._libmx.mxCreateLogicalArray(value.ndim, dim)
-
-            mat_data = self._libmx.mxGetData(pm)
-            np_data = value.real.tostring('F')
-            ctypes.memmove(mat_data, np_data, len(np_data))
-
-        # elif pyvariable.dtype.kind =='S':
-        #     dim = pyvariable.ctypes.shape_as(c_size_t)
-        #     mx = self._libmx.mxCreateNumericArray(c_size_t(pyvariable.ndim),
-        #                                           dim)
-        #     data_old = self._libmx.mxGetData(mx)
-        #     datastring = pyvariable.tostring('F')
-        #     n_datastring = len(datastring)
-        #     memmove(data_old,datastring,n_datastring)
-        # elif pyvariable.dtype.kind =='O':
-        #     raise NotImplementedError('Object arrays are not supported')
-
-        elif isinstance(value, np.ndarray):
-            raise NotImplementedError('Type {} not supported.'.format(value.dtype))
 
+        pm = convert_ndarray_to_mat(self._libmx, value)
 
         self._libeng.engPutVariable(self._ep, name, pm)
 
@@ -560,3 +467,155 @@ class MatlabFunction(object):
         session.eval("clear DOC__")
 
         return doc
+
+
+
+def convert_mat_to_ndarray(libmx, pm):
+    """Convert MATLAB object `pm` to numpy equivalent."""
+
+    ndims = libmx.mxGetNumberOfDimensions(pm)
+    dims = libmx.mxGetDimensions(pm)
+    numelems = libmx.mxGetNumberOfElements(pm)
+    elem_size = libmx.mxGetElementSize(pm)
+    class_name = libmx.mxGetClassName(pm)
+    is_numeric = libmx.mxIsNumeric(pm)
+    is_complex = libmx.mxIsComplex(pm)
+    data = libmx.mxGetData(pm)
+    imag_data = libmx.mxGetImagData(pm)
+
+    if is_numeric:
+        datasize = numelems*elem_size
+
+        real_buffer = ctypes.create_string_buffer(datasize)
+        ctypes.memmove(real_buffer, data, datasize)
+        pyarray = np.ndarray(
+            buffer=real_buffer,
+            shape=dims[:ndims],
+            dtype=class_name,
+            order='F'
+        )
+
+        if is_complex:
+            imag_buffer = ctypes.create_string_buffer(datasize)
+            ctypes.memmove(imag_buffer, imag_data, datasize)
+            pyarray_imag = np.ndarray(
+                buffer=imag_buffer,
+                shape=dims[:ndims],
+                dtype=class_name,
+                order='F'
+            )
+
+            pyarray = pyarray + pyarray_imag * 1j
+
+        out = pyarray.squeeze()
+
+
+    elif class_name == 'char':
+        datasize = numelems + 1
+
+        pystring = ctypes.create_string_buffer(datasize+1)
+        libmx.mxGetString(pm, pystring, datasize)
+
+        out = pystring.value
+
+
+    elif class_name == 'logical':
+        datasize = numelems*elem_size
+
+        buf = ctypes.create_string_buffer(datasize)
+        ctypes.memmove(buf, data, datasize)
+
+        pyarray = np.ndarray(
+            buffer=buf,
+            shape=dims[:ndims],
+            dtype='bool',
+            order='F'
+        )
+
+        out = pyarray.squeeze()
+
+    elif class_name == 'cell':
+        out = []
+        for i in range(numelems):
+            cell = libmx.mxGetCell(pm, i)
+
+            o = convert_mat_to_ndarray(libmx, cell)
+            out.append(o)
+
+        out = np.array(out, dtype='O')
+        out = out.reshape(dims[:ndims], order='F')
+        out = out.squeeze()
+
+    else:
+        raise NotImplementedError('{}-arrays are not supported'.format(class_name))
+
+    return out
+
+
+
+
+def convert_ndarray_to_mat(libmx, arr):
+
+    if isinstance(arr, str):
+        pm = libmx.mxCreateString(arr)
+
+    elif isinstance(arr, dict):
+        raise NotImplementedError('dicts are not supported.')
+
+    else:
+        arr = np.array(arr, ndmin=2)
+
+
+    if isinstance(arr, np.ndarray) and arr.dtype.kind in ['i','u','f','c']:
+        dim = arr.ctypes.shape_as(c_size_t)
+        complex_flag = (arr.dtype.kind == 'c')
+
+        pm = libmx.mxCreateNumericArray(
+            arr.ndim,
+            dim,
+            dtype_to_mat(arr.dtype),
+            complex_flag
+        )
+
+        mat_data = libmx.mxGetData(pm)
+        np_data = arr.real.tostring('F')
+        ctypes.memmove(mat_data, np_data, len(np_data))
+
+        if complex_flag:
+            mat_data = libmx.mxGetImagData(pm)
+            np_data = arr.imag.tostring('F')
+            ctypes.memmove(mat_data, np_data, len(np_data))
+
+    elif isinstance(arr, np.ndarray) and arr.dtype.kind == 'b':
+        dim = arr.ctypes.shape_as(c_size_t)
+
+        pm = libmx.mxCreateLogicalArray(arr.ndim, dim)
+
+        mat_data = libmx.mxGetData(pm)
+        np_data = arr.real.tostring('F')
+        ctypes.memmove(mat_data, np_data, len(np_data))
+
+    elif isinstance(arr, np.ndarray) and arr.dtype.kind == 'O':
+        dim = arr.ctypes.shape_as(c_size_t)
+
+        pm = libmx.mxCreateCellArray(arr.ndim, dim)
+
+        for i,el in enumerate(arr.flatten('F')):
+            p = convert_ndarray_to_mat(libmx, el)
+            libmx.mxSetCell(pm, i, p)
+
+    # elif pyvariable.dtype.kind =='S':
+    #     dim = pyvariable.ctypes.shape_as(c_size_t)
+    #     mx = libmx.mxCreateNumericArray(c_size_t(pyvariable.ndim),
+    #                                           dim)
+    #     data_old = libmx.mxGetData(mx)
+    #     datastring = pyvariable.tostring('F')
+    #     n_datastring = len(datastring)
+    #     memmove(data_old,datastring,n_datastring)
+    # elif pyvariable.dtype.kind =='O':
+    #     raise NotImplementedError('Object arrays are not supported')
+
+    elif isinstance(arr, np.ndarray):
+        raise NotImplementedError('Type {} not supported.'.format(arr.dtype))
+
+    return pm
